REM
REM  Set of microinstructions to be used by the Z80 emulation core
REM  This file is part of the ZilogBoy project
REM  Copyright (C) Mateusz Viste 2010
REM


SUB WriteRegAF(BYREF x AS INTEGER)
  Register.F = (x AND &b11111111)
  Register.A = ((x SHR 8) AND &b11111111)
END SUB
#DEFINE ReadRegAF() (DwordVal(Register.F, Register.A))

SUB WriteRegBC(BYREF x AS INTEGER)
  Register.C = (x AND &b11111111)
  Register.B = ((x SHR 8) AND &b11111111)
END SUB
#DEFINE ReadRegBC() (DwordVal(Register.C, Register.B))

SUB WriteRegDE(BYREF x AS INTEGER)
  Register.E = (x AND &b11111111)
  Register.D = ((x SHR 8) AND &b11111111)
END SUB
#DEFINE ReadRegDE() (DwordVal(Register.E, Register.D))

SUB WriteRegHL(BYREF x AS INTEGER)
  Register.L = (x AND &b11111111)
  Register.H = ((x SHR 8) AND &b11111111)
END SUB
#DEFINE ReadRegHL() (DwordVal(Register.L, Register.H))


SUB SetFlagZ() ' F struct: "ZNHC0000"
  Register.F = (Register.F OR &b10000000)
END SUB
SUB ResetFlagZ()
  Register.F = (Register.F AND &b01111111)
END SUB
#DEFINE GetFlagZ() ((Register.F AND &b10000000) SHR 7)

SUB SetFlagN() ' F struct: "ZNHC0000"
  Register.F = (Register.F OR &b01000000)
END SUB
SUB ResetFlagN()
  Register.F = (Register.F AND &b10111111)
END SUB
#DEFINE GetFlagN() ((Register.F AND &b01000000) SHR 6)

SUB SetFlagH() ' F struct: "ZNHC0000"
  Register.F = (Register.F OR &b00100000)
END SUB
SUB ResetFlagH()
  Register.F = (Register.F AND &b11011111)
END SUB
#DEFINE GetFlagH() ((Register.F AND &b00100000) SHR 5)

SUB SetFlagC() ' F struct: "ZNHC0000"
  Register.F = (Register.F OR &b00010000)
END SUB
SUB ResetFlagC()
  Register.F = (Register.F AND &b11101111)
END SUB
#DEFINE GetFlagC() ((Register.F AND &b00010000) SHR 4)

SUB PushToStack(BYREF TmpRegister AS INTEGER)    ' Push a register to the stack
  MemoryWrite(Register.SP - 1, (TmpRegister AND 255))        ' Write the low byte
  MemoryWrite(Register.SP, ((TmpRegister SHR 8) AND 255))    ' Write the high byte
  Register.SP -= 2  ' decrement SP twice to update the Stack Point address
END SUB

FUNCTION PopFromStack() AS INTEGER
  IF Register.SP < &hFFFE THEN  ' If stack pointer not over available memory...
      Register.SP += 2 ' increments SP twice to update the Stack Point address
      RETURN DwordVal(MemoryRead(Register.SP - 1), MemoryRead(Register.SP))
    ELSE
      RETURN 0
  END IF
END FUNCTION

SUB IncReg8(BYREF RegToInc AS UBYTE) ' Increments by 1 an 8bit register
  ResetFlagN()
  IF RegToInc = 255 THEN
      RegToInc = 0
      SetFlagZ()
    ELSE
      RegToInc += 1
      ResetFlagZ()
  END IF
  IF ((RegToInc AND &b1111) = 0) THEN SetFlagH() ELSE ResetFlagH()
END SUB

SUB DecReg8(BYREF RegToInc AS UBYTE) ' Decrements by 1 an 8bit register
  DIM OldRegToInc AS UBYTE = RegToInc
  SetFlagN()
  IF RegToInc = 0 THEN RegToInc = 255 ELSE RegToInc -= 1
  IF (OldRegToInc AND &b1111) < (RegToInc AND &b1111) THEN SetFlagH() ELSE ResetFlagH()
  IF RegToInc = 0 THEN SetFlagZ() ELSE ResetFlagZ()
END SUB


SUB SubValFromReg8(BYREF RegToSub AS UBYTE, BYREF SubValue AS UBYTE)  ' Subtract a n ubyte value from a 8bit register
  SetFlagN()
  IF RegToSub = SubValue THEN SetFlagZ() ELSE ResetFlagZ()
  IF (RegToSub AND &b1111) < (SubValue AND &b1111) THEN SetFlagH() ELSE ResetFlagH()
  IF RegToSub < SubValue THEN
      RegToSub = 256 - (SubValue - RegToSub)
      SetFlagC()
    ELSE
      RegToSub -= SubValue
      ResetFlagC()
  END IF
END SUB


SUB AddToA(BYREF ValToAdd AS UBYTE) ' Add ValToAdd to register A and sets some flags (8bit arithmetic)
    IF (Register.A AND &b1111) + (ValToAdd AND &b1111) > 15 THEN SetFlagH() ELSE ResetFlagH()
    IF Register.A + ValToADD > 255 THEN SetFlagC() ELSE ResetFlagC()
    Register.A = ((Register.A + ValToAdd) AND &b11111111)
    IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    ResetFlagN()
END SUB


SUB AddToHL(BYREF ValToAdd AS INTEGER)  ' Add ValToAdd to register HL and sets some flags (16bit Z80 arithmetic)
    DIM AS INTEGER HLvalue
    DIM AS INTEGER Result
    HLvalue = ReadRegHL()
    REM I'm not touching H flag, but really not sure what should I do...
    REM The Z80 manual says "Set if carry from bit 11", but NO$GMB doesn't set it :-/
    'IF (HLvalue AND &b111111111111) + (ValToAdd AND &b111111111111) > 4095 THEN SetFlagH() ELSE ResetFlagH()
    IF HLvalue + ValToADD > 65535 THEN SetFlagC() ELSE ResetFlagC()
    WriteRegHL((HLvalue + ValToAdd) AND &b1111111111111111)   ' Write 16bit result into HL
    ResetFlagN()
END SUB


SUB CmpA(BYREF n AS UBYTE)  ' Compare Register.A with n and sets some flags
  IF (Register.A AND 15) < (n AND 15) THEN SetFlagH() ELSE ResetFlagH()
  IF Register.A < n THEN SetFlagC() ELSE ResetFlagC()
  IF Register.A = n THEN SetFlagZ() ELSE ResetFlagZ()
  SetFlagN()
END SUB


SUB XorA(BYREF xorvalz AS UBYTE)
  REM  Xor A with xorvalz, saves results into A, and sets some Z80 flags.
  ResetFlagN() : ResetFlagH() : ResetFlagC()
  Register.A = Register.A XOR xorvalz
  IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
END SUB


SUB AndA(BYREF ValToAndWith AS UBYTE)
  Register.A = Register.A AND ValToAndWith
  IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN() : SetFlagH() : ResetFlagC()
END SUB


SUB OrA(ValToOrWith AS UBYTE)  ' A = A OR ValToOrWith (and sets some Z80 flags)
  Register.A = Register.A OR ValToOrWith
  IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN() : ResetFlagH() : ResetFlagC()
END SUB


SUB SwapUbyte(BYREF ubytevalue AS UBYTE)
  REM  Swaps upper & lower nibbles of ubytevalue and sets some Z80 flags
  ubytevalue = ((ubytevalue AND &b11110000) SHR 4) OR ((ubytevalue AND &b00001111) SHL 4)
  IF ubytevalue = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN() : ResetFlagH() : ResetFlagC()
END SUB


SUB TestBit(BYREF BitNum AS BYTE, BYREF wartosc AS UBYTE)
  REM This tests the BIT BitNum of wartosc and set some Z80 flags
  SELECT CASE BitNum
    CASE 0
      IF (wartosc AND &b00000001) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 1
      IF (wartosc AND &b00000010) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 2
      IF (wartosc AND &b00000100) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 3
      IF (wartosc AND &b00001000) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 4
      IF (wartosc AND &b00010000) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 5
      IF (wartosc AND &b00100000) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 6
      IF (wartosc AND &b01000000) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE 7
      IF (wartosc AND &b10000000) = 0 THEN SetFlagZ() ELSE ResetFlagZ()
    CASE ELSE
      PRINT "Unexpected error! Tried to test invalid bit :-/"
      SLEEP
      END
  END SELECT
  ResetFlagN() : SetFlagH()
END SUB


SUB ShiftRightSRL(BYREF n AS UBYTE)
  REM  Shift n right into Carry (MSB set to 0), and set some flags.
  DIM AS UBYTE Result
  IF (n AND &b00000001) = 0 THEN ResetFlagC() ELSE SetFlagC() ' C contains old bit 0
  ResetFlagN() : ResetFlagH()  ' reset flags N and H
  n = ((n SHR 1) AND &b01111111) ' Shift right by 1 and make sure to reset MSB
  IF n = 0 THEN SetFlagZ() ELSE ResetFlagZ()  ' Set or reset flag Z
END SUB


SUB RotateLeftCarry(BYREF NumRotate AS UBYTE)
  REM  Rotates NumRotate by 1 to the left through carry flag, and sets some Z80 flags.
  REM  This function is used both by RLA AND RL r instructions.
  DIM AS UBYTE TempUbyte
  TempUbyte = (NumRotate AND &b10000000) ' Save old bit 7
  NumRotate = NumRotate SHL 1   ' shift left by 1
  IF NOT GetFlagC() = 0 THEN NumRotate = (NumRotate OR &b00000001)  ' Rotate carry flag into bit 0
  IF TempUbyte = 0 THEN
      ResetFlagC()  ' Restore old bit 7 into carry flag
    ELSE
      SetFlagC()    ' Restore old bit 7 into carry flag
  END IF
  'IF NumRotate = 0 THEN SetFlagZ() ELSE ResetFlagZ()   ' Set or reset the Z flag
  REM I decided to do the same than the NO$GMB emulator: I always reset Z.
  ResetFlagZ()
  ResetFlagN() : ResetFlagH()    ' Flags N & H should be reseted
END SUB


SUB RotateRightCarry(BYREF NumRotate AS UBYTE)
  REM  Rotates NumRotate by 1 to the right through carry flag, and sets some Z80 flags.
  REM  This function is used both by RRA and RR n instructions.
  DIM AS UBYTE TempUbyte
  TempUbyte = (NumRotate AND &b00000001) ' Save old bit 0
  NumRotate = NumRotate SHR 1   ' shift right by 1
  IF NOT GetFlagC() = 0 THEN NumRotate = NumRotate OR &b10000000  ' Rotate carry flag into bit 7
  IF TempUbyte = 0 THEN
      ResetFlagC()  ' Restore old bit 0 into carry flag
    ELSE
      SetFlagC()    ' Restore old bit 0 into carry flag
  END IF
  'IF NumRotate = 0 THEN SetFlagZ() ELSE ResetFlagZ()   ' Set or reset the Z flag
  REM I decided to do the same than the NO$GMB emulator: I always reset Z.
  ResetFlagZ()
  ResetFlagN() : ResetFlagH()    ' Flags N & H should be reseted
END SUB


SUB SbcA(BYREF ValToSub AS UBYTE)
  REM   Subtract content of ValToSub along with the Carry flag from A, and sets some Z80 flags
  DIM AS UBYTE UbyteBuff, FlagC
  IF GetFlagC() = 0 THEN FlagC = 0 ELSE FlagC = 1
  UbyteBuff = ((Register.A - ValToSub - FlagC) AND 255)
  IF Register.A < UbyteBuff THEN SetFlagC() ELSE ResetFlagC()
  IF (Register.A AND &b1111) < (UbyteBuff AND &b1111) THEN SetFlagH() ELSE ResetFlagH()
  Register.A = UbyteBuff
  SetFlagN()
  IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
END SUB


SUB AdcA(BYREF ValToAdd AS UBYTE)
  REM Add content of ValToAdd along with Carry flag to A, and sets some Z80 flags
  DIM AS UBYTE UbyteBuff, FlagC
  IF GetFlagC() = 0 THEN FlagC = 0 ELSE FlagC = 1
  UbyteBuff = ((Register.A + ValToAdd + FlagC) AND 255)
  IF UbyteBuff = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN()
  IF Register.A + ValToAdd + FlagC > 255 THEN SetFlagC() ELSE ResetFlagC() ' Set if carry from bit 7
  IF (Register.A AND &b1111) + (ValToAdd AND &b1111) + FlagC > 15 THEN SetFlagH() ELSE ResetFlagH() ' Set if carry from bit 3
  Register.A = UbyteBuff
END SUB


SUB ShiftSLA(BYREF ValueToShift AS UBYTE)
  REM  Shift ValueToShift left by 1 bit, putting old bit 7 into Carry flag.
  IF (ValueToShift AND &b10000000) = 0 THEN ResetFlagC() ELSE SetFlagC()
  ValueToShift = ValueToShift SHL 1
  IF ValueToShift = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN() : ResetFlagH()
END SUB


SUB ShiftSRA(BYREF ValueToShift AS UBYTE)
  REM  Shift ValueToShift right by 1 bit, putting old bit 0 into Carry flag. MSB doesn't change.
  IF (ValueToShift AND &b00000001) = 0 THEN ResetFlagC() ELSE SetFlagC()
  ValueToShift = ValueToShift SHR 1
  IF (ValueToShift AND &b01000000) = 0 THEN ' Restore old bit 7 (MSB) as it was before
      ValueToShift = ValueToShift AND &b01111111  ' reset MSB
    ELSE
      ValueToShift = ValueToShift OR &b10000000   ' set MSB
  END IF
  IF ValueToShift = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagN() : ResetFlagH()
END SUB


SUB RRCA()
  REM  Rotates A by 1 to the right, and sets some Z80 flags.
  DIM AS UBYTE TempUbyte
  TempUbyte = (Register.A AND &b00000001) ' Save old bit 7
  Register.A = Register.A SHR 1   ' shift right by 1
  IF TempUbyte = 0 THEN  ' if old bit 7 was 0
      ResetFlagC()  ' Reset carry flag
      Register.A = (Register.A AND &b01111111)  ' restore old bit 7 into bit 0
    ELSE  ' if old bit 7 was 1
      SetFlagC()    ' Set carry flag
      Register.A = Register.A OR &b10000000  ' restore old bit 7 into bit 0
  END IF
  REM I *do* touch the Z flag. The Z80 manual tells to not touch it (while the Gameboy
  REM CPU manual tells that it should be reset if result is 0...)
  REM I decided to do the same than the NO$GMB emulator: I always reset Z.
  'IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()   ' Set or reset the Z flag
  ResetFlagZ()
  ResetFlagN() : ResetFlagH()    ' Flags N & H should be reseted
END SUB


SUB RLCA()
  REM  Rotates NumRotate by 1 to the left, and sets some Z80 flags.
  DIM AS UBYTE TempUbyte
  TempUbyte = (Register.A AND &b10000000) ' Save old bit 7
  Register.A = Register.A SHL 1   ' shift left by 1
  IF TempUbyte = 0 THEN
      ResetFlagC()  ' Reset carry flag
      Register.A = (Register.A AND &b11111110)  ' restore old bit 7 into bit 0
    ELSE
      SetFlagC()    ' Set carry flag
      Register.A = Register.A OR &b00000001  ' restore old bit 7 into bit 0
  END IF
  REM I *do* touch the Z flag. The Z80 manual tells to not touch it (while the Gameboy
  REM CPU manual tells that it should be reset if result is 0...)
  REM I decided to do the same than the NO$GMB emulator: I always reset Z.
  'IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()   ' Set or reset the Z flag
  ResetFlagZ()
  ResetFlagN() : ResetFlagH()    ' Flags N & H should be reseted
END SUB


SUB RotateRLC(BYREF valtorot AS UBYTE)
  REM The contents of register r are rotated left 1-bit position.
  REM The content of bit 7 is copied to the Carry flag and also to bit 0.
  IF (valtorot AND &b10000000) = 0 THEN ResetFlagC() ELSE SetFlagC()  ' save bit7 into flag C
  valtorot = valtorot SHL 1
  REM now restore old bit 7 (holded in flag C) into bit 0
  IF GetFlagC() = 0 THEN valtorot = valtorot AND &b11111110 ELSE valtorot = valtorot OR &b00000001
  IF valtorot = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagH() : ResetFlagN()
END SUB


SUB RotateRRC(BYREF valtorot AS UBYTE)
  REM The contents of register r are rotated right 1-bit position.
  REM The content of bit 0 is copied to the Carry flag and also to bit 7.
  IF (valtorot AND &b00000001) = 0 THEN ResetFlagC() ELSE SetFlagC()  ' save bit0 into flag C
  valtorot = valtorot SHR 1
  REM now restore old bit 0 (holded in flag C) into bit 7
  IF GetFlagC() = 0 THEN valtorot = valtorot AND &b01111111 ELSE valtorot = valtorot OR &b10000000
  IF valtorot = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagH() : ResetFlagN()
END SUB


SUB AdjustDAA() ' Decimal adjust on register A, and sets some Z80 flags
  DIM AS UBYTE FlagsNCH = &b111
  IF GetFlagN() = 0 THEN FlagsNCH = (FlagsNCH AND &b011)
  IF GetFlagC() = 0 THEN FlagsNCH = (FlagsNCH AND &b101)
  IF GetFlagH() = 0 THEN FlagsNCH = (FlagsNCH AND &b110)
  SELECT CASE FlagsNCH
    CASE &b000 ' no flags
      IF ((Register.A AND &b00001111) < 10) AND (((Register.A AND &b11110000) SHR 4) < 10) THEN
          Register.A += &h00
          ResetFlagC()
        ELSEIF ((Register.A AND &b00001111) < 9) AND (((Register.A AND &b11110000) SHR 4) > 9) THEN
          Register.A += &h06
          ResetFlagC()
        ELSEIF ((Register.A AND &b00001111) > 9) AND (((Register.A AND &b11110000) SHR 4) < 10) THEN
          Register.A += &h60
          SetFlagC()
        ELSEIF ((Register.A AND &b00001111) > 8) AND (((Register.A AND &b11110000) SHR 4) > 9) THEN
          Register.A += &h66
          SetFlagC()
      END IF
    CASE &b010 ' Carry
      IF ((Register.A AND &b00001111) < 3) AND (((Register.A AND &b11110000) SHR 4) < 10) THEN
          Register.A += &h60
          SetFlagC()
        ELSEIF ((Register.A AND &b00001111) < 3) AND (((Register.A AND &b11110000) SHR 4) > 9) THEN
          Register.A += &h66
          SetFlagC()
      END IF
    CASE &b100 ' Subtract
      IF ((Register.A AND &b00001111) < 10) AND (((Register.A AND &b11110000) SHR 4) < 10) THEN
        Register.A += &h00
        ResetFlagC()
      END IF
    CASE &b110 ' Subtract + Carry
      IF ((Register.A AND &b00001111) > 6) AND (((Register.A AND &b11110000) SHR 4) < 10) THEN
        Register.A += &hA0
        SetFlagC()
      END IF
    CASE &b001 ' Half-carry
      IF ((Register.A AND &b00001111) < 10) AND (((Register.A AND &b11110000) SHR 4) < 4) THEN
          Register.A += &h06
          ResetFlagC()
        ELSEIF ((Register.A AND &b00001111) > 9) AND (((Register.A AND &b11110000) SHR 4) < 4) THEN
          Register.A += &h66
          SetFlagC()
      END IF
    CASE &b011 ' Half-carry + Carry
      IF ((Register.A AND &b00001111) < 4) AND (((Register.A AND &b11110000) SHR 4) < 4) THEN
        Register.A += &h66
        SetFlagC()
      END IF
    CASE &b101 ' Half-carry + Subtract
      IF ((Register.A AND &b00001111) < 9) AND (((Register.A AND &b11110000) SHR 4) > 5) THEN
        Register.A += &hFA
        ResetFlagC()
      END IF
    CASE &b111 ' Half-carry + Subtract + Carry
      IF ((Register.A AND &b00001111) > 5) AND ((Register.A AND &b00001111) < 8) AND (((Register.A AND &b11110000) SHR 4) > 5) THEN
        Register.A += &h9A
        SetFlagC()
      END IF
  END SELECT
  IF Register.A = 0 THEN SetFlagZ() ELSE ResetFlagZ()
  ResetFlagH()
END SUB


SUB MemoryWrite16(BYREF MemAddr AS UINTEGER, BYREF DwordData AS UINTEGER)
  MemoryWrite(DwordVal(MemoryRead(Register.PC + 1),MemoryRead(Register.PC + 2)), (DwordData AND &b0000000011111111))
  MemoryWrite(DwordVal(MemoryRead(Register.PC + 1),MemoryRead(Register.PC + 2)) + 1, (DwordData AND &b1111111100000000) SHR 8)
END SUB


SUB CpuPowerUp()
  DIM AS INTEGER x
  Register.PC = &h100   ' The execution of code starts at 0x100
  WriteRegBC(&h13)
  WriteRegDE(&hD8)
  WriteRegHL(&h14D)
  Register.SP = &hFFFE  ' Stack pointer is inited at &hFFFE
  MemoryWrite(&hFF05, &h00)   ' TIMA
  MemoryWrite(&hFF06, &h00)   ' TMA
  MemoryWrite(&hFF07, &h00)   ' TAC
  MemoryWrite(&hFF10, &h80)   ' NR10
  MemoryWrite(&hFF11, &hBF)   ' NR11
  MemoryWrite(&hFF12, &hF3)   ' NR12
  MemoryWrite(&hFF14, &hBF)   ' NR14
  MemoryWrite(&hFF16, &h3F)   ' NR21
  MemoryWrite(&hFF17, &h00)   ' NR22
  MemoryWrite(&hFF19, &hBF)   ' NR24
  MemoryWrite(&hFF1A, &h7F)   ' NR30
  MemoryWrite(&hFF1B, &hFF)   ' NR31
  MemoryWrite(&hFF1C, &h9F)   ' NR32
  MemoryWrite(&hFF1E, &hBF)   ' NR33
  MemoryWrite(&hFF20, &hFF)   ' NR41
  MemoryWrite(&hFF21, &h00)   ' NR42
  MemoryWrite(&hFF22, &h00)   ' NR43
  MemoryWrite(&hFF23, &hBF)   ' NR30
  MemoryWrite(&hFF24, &h77)   ' NR50
  MemoryWrite(&hFF25, &hF3)   ' NR51
  MemoryWrite(&hFF26, &hF1)   ' NR52  / On GB only
  'MemoryWrite(&hFF26, &hF0)   ' NR52  / On SGB only
  MemoryWrite(&hFF40, &h91)   ' LCDC
  MemoryWrite(&hFF42, &h00)   ' SCY
  MemoryWrite(&hFF43, &h00)   ' SCX
  MemoryWrite(&hFF45, &h00)   ' LYC
  MemoryWrite(&hFF47, &hFC)   ' BGP
  MemoryWrite(&hFF48, &hFF)   ' OBP0
  MemoryWrite(&hFF49, &hFF)   ' OBP1
  MemoryWrite(&hFF4A, &h00)   ' WY
  MemoryWrite(&hFF4B, &h00)   ' WX
  MemoryWrite(&hFFFF, &h00)   ' IE
  SetFlagZ()      ' Set some flags
  SetFlagH()      ' - as done by
  SetFlagC()      ' the NO$CASH
  Register.A = 1  ' GB emulator.
  REM Now let's load all timer values (time for every instruction to be executed)
  cpc(&h00) = 1: cpc(&H01) = 3: cpc(&H02) = 2: cpc(&H03) = 2
  cpc(&h04) = 1: cpc(&H05) = 1: cpc(&H06) = 2: cpc(&H07) = 1
  cpc(&h08) = 5: cpc(&H09) = 2: cpc(&H0A) = 2: cpc(&H0B) = 2
  cpc(&h0C) = 1: cpc(&H0D) = 1: cpc(&H0E) = 2: cpc(&H0F) = 1
  cpc(&h10) = 1: cpc(&H11) = 3: cpc(&H12) = 2: cpc(&H13) = 2
  cpc(&h14) = 1: cpc(&H15) = 1: cpc(&H16) = 2: cpc(&H17) = 1
  cpc(&h18) = 2: cpc(&H19) = 2: cpc(&H1A) = 2: cpc(&H1B) = 2
  cpc(&h1C) = 1: cpc(&H1D) = 1: cpc(&H1E) = 2: cpc(&H1F) = 1
  cpc(&h20) = 2: cpc(&H21) = 3: cpc(&H22) = 2: cpc(&H23) = 2
  cpc(&h24) = 1: cpc(&H25) = 1: cpc(&H26) = 2: cpc(&H27) = 1
  cpc(&h28) = 2: cpc(&H29) = 2: cpc(&H2A) = 2: cpc(&H2B) = 2
  cpc(&h2C) = 1: cpc(&H2D) = 1: cpc(&H2E) = 2: cpc(&H2F) = 1
  cpc(&h30) = 2: cpc(&H31) = 3: cpc(&H32) = 2: cpc(&H33) = 2
  cpc(&h34) = 3: cpc(&H35) = 3: cpc(&H36) = 3: cpc(&H37) = 1
  cpc(&h38) = 2: cpc(&H39) = 2: cpc(&H3A) = 2: cpc(&H3B) = 2
  cpc(&h3C) = 1: cpc(&H3D) = 1: cpc(&H3E) = 2: cpc(&H3F) = 1
  cpc(&h40) = 1: cpc(&H41) = 1: cpc(&H42) = 1: cpc(&H43) = 1
  cpc(&h44) = 1: cpc(&H45) = 1: cpc(&H46) = 2: cpc(&H47) = 1
  cpc(&h48) = 1: cpc(&H49) = 1: cpc(&H4A) = 1: cpc(&H4B) = 1
  cpc(&h4C) = 1: cpc(&H4D) = 1: cpc(&H4E) = 2: cpc(&H4F) = 1
  cpc(&h50) = 1: cpc(&H51) = 1: cpc(&H52) = 1: cpc(&H53) = 1
  cpc(&h54) = 1: cpc(&H55) = 1: cpc(&H56) = 2: cpc(&H57) = 1
  cpc(&h58) = 1: cpc(&H59) = 1: cpc(&H5A) = 1: cpc(&H5B) = 1
  cpc(&h5C) = 1: cpc(&H5D) = 1: cpc(&H5E) = 2: cpc(&H5F) = 1
  cpc(&h60) = 1: cpc(&H61) = 1: cpc(&H62) = 1: cpc(&H63) = 1
  cpc(&h64) = 1: cpc(&H65) = 1: cpc(&H66) = 2: cpc(&H67) = 1
  cpc(&h68) = 1: cpc(&H69) = 1: cpc(&H6A) = 1: cpc(&H6B) = 1
  cpc(&h6C) = 1: cpc(&H6D) = 1: cpc(&H6E) = 2: cpc(&H6F) = 1
  cpc(&h70) = 2: cpc(&H71) = 2: cpc(&H72) = 2: cpc(&H73) = 2
  cpc(&h74) = 2: cpc(&H75) = 2: cpc(&H76) = 1: cpc(&H77) = 2
  cpc(&h78) = 1: cpc(&H79) = 1: cpc(&H7A) = 1: cpc(&H7B) = 1
  cpc(&h7C) = 1: cpc(&H7D) = 1: cpc(&H7E) = 2: cpc(&H7F) = 1
  cpc(&h80) = 1: cpc(&H81) = 1: cpc(&H82) = 1: cpc(&H83) = 1
  cpc(&h84) = 1: cpc(&H85) = 1: cpc(&H86) = 2: cpc(&H87) = 1
  cpc(&h88) = 1: cpc(&H89) = 1: cpc(&H8A) = 1: cpc(&H8B) = 1
  cpc(&h8C) = 1: cpc(&H8D) = 1: cpc(&H8E) = 2: cpc(&H8F) = 1
  cpc(&h90) = 1: cpc(&H91) = 1: cpc(&H92) = 1: cpc(&H93) = 1
  cpc(&h94) = 1: cpc(&H95) = 1: cpc(&H96) = 2: cpc(&H97) = 1
  cpc(&h98) = 1: cpc(&H99) = 1: cpc(&H9A) = 1: cpc(&H9B) = 1
  cpc(&h9C) = 1: cpc(&H9D) = 1: cpc(&H9E) = 2: cpc(&H9F) = 1
  cpc(&hA0) = 1: cpc(&HA1) = 1: cpc(&HA2) = 1: cpc(&HA3) = 1
  cpc(&hA4) = 1: cpc(&HA5) = 1: cpc(&HA6) = 2: cpc(&HA7) = 1
  cpc(&hA8) = 1: cpc(&HA9) = 1: cpc(&HAA) = 1: cpc(&HAB) = 1
  cpc(&hAC) = 1: cpc(&HAD) = 1: cpc(&HAE) = 2: cpc(&HAF) = 1
  cpc(&hB0) = 1: cpc(&HB1) = 1: cpc(&HB2) = 1: cpc(&HB3) = 1
  cpc(&hB4) = 1: cpc(&HB5) = 1: cpc(&HB6) = 2: cpc(&HB7) = 1
  cpc(&hB8) = 1: cpc(&HB9) = 1: cpc(&HBA) = 1: cpc(&HBB) = 1
  cpc(&hBC) = 1: cpc(&HBD) = 1: cpc(&HBE) = 2: cpc(&HBF) = 1
  cpc(&hC0) = 2: cpc(&HC1) = 3: cpc(&HC2) = 3: cpc(&HC3) = 4
  cpc(&hC4) = 3: cpc(&HC5) = 4: cpc(&HC6) = 2: cpc(&HC7) = 4
  cpc(&hC8) = 2: cpc(&HC9) = 4: cpc(&HCA) = 3: cpc(&HCB) = 2
  cpc(&hCC) = 3: cpc(&HCD) = 6: cpc(&HCE) = 2: cpc(&HCF) = 4
  cpc(&hD0) = 2: cpc(&HD1) = 3: cpc(&HD2) = 3: cpc(&HD3) = 3
  cpc(&hD4) = 3: cpc(&HD5) = 4: cpc(&HD6) = 2: cpc(&HD7) = 4
  cpc(&hD8) = 2: cpc(&HD9) = 4: cpc(&HDA) = 3: cpc(&HDB) = 0
  cpc(&hDC) = 3: cpc(&HDD) = 0: cpc(&HDE) = 2: cpc(&HDF) = 4
  cpc(&hE0) = 3: cpc(&HE1) = 3: cpc(&HE2) = 2: cpc(&HE3) = 0
  cpc(&hE4) = 0: cpc(&HE5) = 4: cpc(&HE6) = 2: cpc(&HE7) = 4
  cpc(&hE8) = 4: cpc(&HE9) = 1: cpc(&HEA) = 4: cpc(&HEB) = 0
  cpc(&hEC) = 0: cpc(&HED) = 0: cpc(&HEE) = 2: cpc(&HEF) = 4
  cpc(&hF0) = 2: cpc(&HF1) = 3: cpc(&HF2) = 2: cpc(&HF3) = 1
  cpc(&hF4) = 0: cpc(&HF5) = 4: cpc(&HF6) = 2: cpc(&HF7) = 4
  cpc(&hF8) = 2: cpc(&HF9) = 2: cpc(&HFA) = 4: cpc(&HFB) = 1
  cpc(&hFC) = 0: cpc(&HFD) = 0: cpc(&HFE) = 2: cpc(&HFF) = 4
  FOR x = 0 TO 255
    cpc(x) = cpc(x) * 4
    IF cpc(x) = 0 THEN cpc(x) = 1
  NEXT x
  EmulationStartTimer = TIMER    ' Save timer for later timing adjustements
END SUB
