REM
REM  Part of the ZilogBoy project
REM

FUNCTION CheckChecksums(BYREF RomSize AS UINTEGER) AS UBYTE
  DIM AS INTEGER x, CmpChecksum, TotalChecksum
  DIM AS UBYTE tByteBuff, Result = 0
  REM Returns 0 if GameBoy checksums are okay, non-zero otherwise.
  REM 14Dh: A complement check is computed as the sum of the bytes in the header of the program subtracted by one and then two's complemented. The complement check is required by the NINTENDO GAMEBOY device.
  FOR x = &h134 TO &h14C
    tByteBuff = MemoryROM(x)
    CmpChecksum = (CmpChecksum + tByteBuff) AND &b11111111
  NEXT x
  CmpChecksum = (231 - CmpChecksum)
  IF CmpChecksum < 0 THEN  ' If negative, get the 2-complement
    CmpChecksum = ABS(CmpChecksum)
    CmpChecksum = (CmpChecksum XOR &b11111111)
    CmpChecksum += 1
  END IF
  IF MemoryROM(&h14D) = CmpChecksum THEN
      REM 014E-014F: Checksum (higher byte first) produced by adding all bytes of a cartridge except for two checksum bytes and taking two lower bytes of the result. (GameBoy ignores this value.)
      FOR x = 0 TO RomSize - 1
        tByteBuff = MemoryROM(x)
        IF (x <> &h14E) AND (x <> &h14F) THEN TotalChecksum += tByteBuff
        TotalChecksum = (TotalChecksum AND &b1111111111111111) ' Keep the checksum 32bit wide
      NEXT x
      tByteBuff = MemoryROM(&h14E)
      IF tByteBuff <> ((TotalChecksum SHR 8) AND &b11111111) THEN Result = 3 ' Check first byte of checksum
      tByteBuff = MemoryROM(&h14F)
      IF tByteBuff <> (TotalChecksum AND &b11111111) THEN Result = 2         ' Check second byte of checksum
    ELSE
      Result = 1
  END IF
  RETURN Result
END FUNCTION


SUB LoadRom(RomFile AS STRING)
  DIM NintendoSignature(&h104 TO &h133) AS UBYTE => {&hCE,&hED,&h66,&h66,&hCC,&h0D,&h00,&h0B,_
                                                     &h03,&h73,&h00,&h83,&h00,&h0C,&h00,&h0D,_
                                                     &h00,&h08,&h11,&h1F,&h88,&h89,&h00,&h0E,_
                                                     &hDC,&hCC,&h6E,&hE6,&hDD,&hDD,&hD9,&h99,_
                                                     &hBB,&hBB,&h67,&h63,&h6E,&h0E,&hEC,&hCC,_
                                                     &hDD,&hDC,&h99,&h9F,&hBB,&hB9,&h33,&h3E}
  DIM AS INTEGER RomHandler, x, OpenResult
  DIM AS UINTEGER RomSize
  RomHandler = FREEFILE
  IF FileExists(RomFile) = 0 THEN
      PRINT "Error: The specified ROM file doesn't exists!"
      END
    ELSE
      OpenResult = OPEN(RomFile, FOR BINARY, AS #RomHandler)
      IF OpenResult = 0 THEN
        RomSize = LOF(RomHandler)
        IF RomSize < 16384 THEN
            CLOSE #RomHandler
            PRINT "Error: The file is too small to be a valid GameBoy ROM!"
            END
          ELSE
            FOR x = 0 TO RomSize - 1
              GET #RomHandler, x + 1, MemoryROM(x)
            NEXT x
            CLOSE #RomHandler
          END IF
        ELSE
          PRINT "Error: The specified ROM file could not be opened!"
          END
      END IF
  END IF
  REM  Check for ROM signature
  FOR x = &h104 TO &h133
    IF MemoryROM(x) <> NintendoSignature(x) THEN
      PRINT "Error: The file is not a valid GameBoy ROM!"
      END
    END IF
  NEXT x
  SELECT CASE CheckChecksums(RomSize)
    CASE 0
      REM Nothing here, everything is ok.
    CASE 1
      PRINT "Warning: This ROM has wrong checksum values. It is probably a corrupted ROM!"
      SLEEP 500, 1
    CASE 2, 3
      PRINT "Warning: This ROM has wrong checksum values. It might be a corrupted ROM!"
      SLEEP 500, 1
    CASE ELSE
      PRINT "Error: An unexpected condition occurred when checking ROM's checksums!"
      END
  END SELECT
END SUB
