REM
REM  Video emulation core of a GameBoy device
REM  This file is part of the ZilogBoy project
REM  Copyright (C) Mateusz Viste 2010
REM

 ' FF40 (LCDC register):
 ' Bit7 0=Display Off  1=Display On
 ' Bit6 0=TileMap at 9800-9BFF   1=TileMap at 9C00-9FFF
 ' Bit5 0=WindowDisplay disabled  1=WindowDisplay enabled
 ' Bit4 0=BG&Window tile data at 8800-97FF   1=BG&Window tile data at 8000-8FFF
 ' Bit3 0=BgTile Map display at 9800-9BFF   1=BgTitle Map display at 9C00-9FFF
 ' Bit2 0=SpriteSize=8x8    1=SpriteSize=8x16
 ' Bit1 0=Do not display Sprites   1=Display sprites
 ' Bit0 0=Do not display background   1=Display background

 ' Generating an interrupt means "enabling the corresponding bit of the IF register (FF0F)"

DIM SHARED AS UINTEGER VideoClkCounterMode, VideoClkCounterVBlank
DIM SHARED ScreenBuffer(0 TO 255, 0 TO 255) AS UBYTE
DIM SHARED ScreenPalette(0 TO 3) AS UINTEGER  ' Palette of shades to use onscreen
ScreenPalette(0) = RGB(&hFA,&hFA,&hFA)  ' Brightest shade
ScreenPalette(1) = RGB(&hAF,&hAF,&hAF)
ScreenPalette(2) = RGB(&h64,&h64,&h64)
ScreenPalette(3) = RGB(&h19,&h19,&h19)  ' Darkest shade


FUNCTION GbPalette(BYREF ColIdx AS UBYTE, BYREF PalType AS UBYTE) AS UBYTE
  REM  Returns the RGB value of a color index from one of the palettes
  REM   ColIdx: 0..3
  REM   PalType: 0=BGP  1=OBP0  2=OBP1
  DIM AS UBYTE ColValue
  SELECT CASE ColIdx
    CASE 0
      ColValue = (IoRegisters(&hFF47 + PalType) AND &b00000011)
    CASE 2
      ColValue = ((IoRegisters(&hFF47 + PalType) AND &b00001100) SHR 2)
    CASE 1
      ColValue = ((IoRegisters(&hFF47 + PalType) AND &b00110000) SHR 4)
    CASE 3
      ColValue = ((IoRegisters(&hFF47 + PalType) AND &b11000000) SHR 6)
  END SELECT
  RETURN ColValue
END FUNCTION


SUB DrawBackground()
  DIM AS INTEGER TilesDataAddr, BgTilesMapAddr, WinTilesMapAddr, TileNum, TileToDisplay, PixelX, PixelY
  DIM AS UBYTE LCDC, x, y, z, t, UbyteBuff1, UbyteBuff2
  DIM TileBuffer(0 TO 63) AS UBYTE
  LCDC = IoRegisters(&hFF40)
  IF (LCDC AND &b00000001) = 1 THEN  ' If "BackgroundEnabled" bit is set then draw background
    REM Get starting address of tiles data
    IF (LCDC AND &b00010000) = 0 THEN
        TilesDataAddr = &h8800
        'TilesDataAddr = &h87FF
      ELSE
        TilesDataAddr = &h8000
    END IF
    IF (LCDC AND &b00001000) = 0 THEN ' Get address of BGP tiles *map*
        BgTilesMapAddr = &h9800
      ELSE
        BgTilesMapAddr = &h9C00
    END IF
    IF (LCDC AND &b01000000) = 0 THEN WinTilesMapAddr = &h9800 ELSE WinTilesMapAddr = &h9C00
    FOR y = 0 TO 31
      FOR x = 0 TO 31
        TileNum = (y * 32) + x
        IF TilesDataAddr = &h8000 THEN
            TileToDisplay = VideoRAM(BgTilesMapAddr + TileNum)
          ELSE
            TileToDisplay = UbyteToByte(VideoRAM(BgTilesMapAddr + TileNum)) + 128
        END IF
        FOR z = 0 TO 15 STEP 2
          UbyteBuff1 = VideoRAM(TilesDataAddr + (TileToDisplay * 16) + z)
          UbyteBuff2 = VideoRAM(TilesDataAddr + (TileToDisplay * 16) + z + 1)
          TileBuffer((z * 4) + 7) =  ((UbyteBuff2 AND &b00000001) OR ((UbyteBuff1 AND &b00000001) SHL 1))
          TileBuffer((z * 4) + 6) = (((UbyteBuff2 AND &b00000010) SHR 1) OR (((UbyteBuff1 AND &b00000010) SHR 1) SHL 1))
          TileBuffer((z * 4) + 5) = (((UbyteBuff2 AND &b00000100) SHR 2) OR (((UbyteBuff1 AND &b00000100) SHR 2) SHL 1))
          TileBuffer((z * 4) + 4) = (((UbyteBuff2 AND &b00001000) SHR 3) OR (((UbyteBuff1 AND &b00001000) SHR 3) SHL 1))
          TileBuffer((z * 4) + 3) = (((UbyteBuff2 AND &b00010000) SHR 4) OR (((UbyteBuff1 AND &b00010000) SHR 4) SHL 1))
          TileBuffer((z * 4) + 2) = (((UbyteBuff2 AND &b00100000) SHR 5) OR (((UbyteBuff1 AND &b00100000) SHR 5) SHL 1))
          TileBuffer((z * 4) + 1) = (((UbyteBuff2 AND &b01000000) SHR 6) OR (((UbyteBuff1 AND &b01000000) SHR 6) SHL 1))
          TileBuffer((z * 4) + 0) = (((UbyteBuff2 AND &b10000000) SHR 7) OR (((UbyteBuff1 AND &b10000000) SHR 7) SHL 1))
        NEXT z
        FOR z = 0 TO 7
          FOR t = 0 TO 7
            PixelX = ((x * 8) + t - IoRegisters(&hFF43))  ' &hFF43 is the SCX register
            PixelY = ((y * 8) + z - IoRegisters(&hFF42))  ' &hFF42 is the SCY register
            IF PixelX < 0 THEN PixelX += 256
            IF PixelY < 0 THEN PixelY += 256
            ScreenBuffer(PixelX, PixelY) = GbPalette(TileBuffer((z * 8) + t), 0)
          NEXT t
        NEXT z
      NEXT x
    NEXT y
  END IF
END SUB


SUB DrawSprites()
  DIM AS UBYTE SpritePosX, SpritePosY, PatternNum, SpriteFlags, UbyteBuff1, UbyteBuff2, LCDC, SpritePalette
  DIM AS INTEGER x, y, z, t
  DIM SpriteBuff(1 TO 64) AS BYTE
  LCDC = IoRegisters(&hFF40)
  FOR x = 0 TO 39  ' Read 40 blocks of 4 bytes starting from &hFE00
    SpritePosY = SpriteOAM(&hFE00 + (x * 4) + 0) - 16   ' sprite's ypos on screen
    SpritePosX = SpriteOAM(&hFE00 + (x * 4) + 1) - 8    ' sprite's xpos on screen
    IF LCDC AND &b00000100 = 0 THEN  ' If Bit 2 of FF40 is reset -> sprites are 8x8
        PatternNum = SpriteOAM(&hFE00 + (x * 4) + 2)    ' pattern num
      ELSE  ' Else sprites are 16x8 (LSB ignored in 8x16 sprite mode)
        PatternNum = SpriteOAM(&hFE00 + (x * 4) + 2) AND &b11111110
    END IF
    SpriteFlags = SpriteOAM(&hFE00 + (x * 4) + 3)        ' Flags
    IF SpriteFlags AND &b00010000 = 0 THEN SpritePalette = 1 ELSE SpritePalette = 2
    IF SpritePosX >= 8 AND SpritePosX <= 255 AND SpritePosY >= 16 AND SpritePosY <= 255 THEN
        FOR z = 0 TO 15 STEP 2
          UbyteBuff1 = VideoRAM(&h8000 + (PatternNum * 16) + z)
          UbyteBuff2 = VideoRAM(&h8000 + (PatternNum * 16) + z + 1)
          SpriteBuff((z * 4) + 7) =  ((UbyteBuff2 AND &b00000001) OR ((UbyteBuff1 AND &b00000001) SHL 1))
          SpriteBuff((z * 4) + 6) = (((UbyteBuff2 AND &b00000010) SHR 1) OR (((UbyteBuff1 AND &b00000010) SHR 1) SHL 1))
          SpriteBuff((z * 4) + 5) = (((UbyteBuff2 AND &b00000100) SHR 2) OR (((UbyteBuff1 AND &b00000100) SHR 2) SHL 1))
          SpriteBuff((z * 4) + 4) = (((UbyteBuff2 AND &b00001000) SHR 3) OR (((UbyteBuff1 AND &b00001000) SHR 3) SHL 1))
          SpriteBuff((z * 4) + 3) = (((UbyteBuff2 AND &b00010000) SHR 4) OR (((UbyteBuff1 AND &b00010000) SHR 4) SHL 1))
          SpriteBuff((z * 4) + 2) = (((UbyteBuff2 AND &b00100000) SHR 5) OR (((UbyteBuff1 AND &b00100000) SHR 5) SHL 1))
          SpriteBuff((z * 4) + 1) = (((UbyteBuff2 AND &b01000000) SHR 6) OR (((UbyteBuff1 AND &b01000000) SHR 6) SHL 1))
          SpriteBuff((z * 4) + 0) = (((UbyteBuff2 AND &b10000000) SHR 7) OR (((UbyteBuff1 AND &b10000000) SHR 7) SHL 1))
        NEXT z
        FOR z = 0 TO 7
          FOR t = 0 TO 7
            REM  Update all pixels, but not 0 (which are transparent for a sprite)
            IF SpriteBuff((z * 8) + t) > 0 THEN ScreenBuffer(SpritePosX + t, SpritePosY + z) = GbPalette(SpriteBuff((z * 8) + t), SpritePalette)
          NEXT t
        NEXT z
    END IF
  NEXT x
END SUB


SUB RefreshScreen()
  DIM AS UINTEGER x, y, a, b, c, d
  DIM AS UBYTE ScaleStepX, ScaleStepY
  DIM AS UINTEGER PTR sPtr = Screenptr
  SCREENLOCK()
  FOR ScaleStepY = 0 TO (GraphicScaleFactor - 1)
    a = GraphicWidth * ScaleStepY
    FOR y = 0 TO 143
      b = y * GraphicWidth * GraphicScaleFactor
      FOR x = 0 TO 159 ' 255 for debug
        c = x * GraphicScaleFactor
        FOR ScaleStepX = 0 TO (GraphicScaleFactor - 1)                      ' it's *much* faster to
          sPtr[b + a + c + ScaleStepX] = ScreenPalette(ScreenBuffer(x, y))  ' write into ScreenPTR
        NEXT ScaleStepX                                                     ' than rely on PSET.
      NEXT x
    NEXT y
  NEXT ScaleStepY
  SCREENUNLOCK()
END SUB


SUB VideoSysUpdate(cycles AS UINTEGER)
  DIM AS UBYTE NewLY
  VideoClkCounterVBlank += cycles
  IF VideoClkCounterVBlank >= 70224 THEN
    VideoClkCounterVBlank -= 70224
  END IF

  REM  MemoryInternalHiRAM(&hFF44) is the LY register (current vertical scanline)
  NewLY = VideoClkCounterVBlank \ 458          ' LY should be between 0..153
  IF IoRegisters(&hFF44) <> NewLY THEN         ' If LY changed, then...
    IoRegisters(&hFF44) = NewLY                ' Save new LY
    'PrintScanline(IoRegisters(&hFF44))         ' Write the current scanline to screen
  END IF
  IF IoRegisters(&hFF44) = IoRegisters(&hFF45) THEN       ' Compare LY with LYC, and if equal,
      IoRegisters(&hFF41) = IoRegisters(&hFF41) OR &b100  ' set the coincidence (2) bit of the
      IF IoRegisters(&hFF41) AND &b01000000 > 0 THEN      ' STAT register (FF41) and (if enabled
        IoRegisters(&hFF0F) = IoRegisters(&hFF0F) OR &b10 ' via bit 6 of STAT register) trigger
      END IF                                              ' the Coincidence (LCDC) interrupt.
    ELSE
      IoRegisters(&hFF41) = IoRegisters(&hFF41) AND &b11111011 ' reset the coincidence flag
  END IF

  IF IoRegisters(&hFF44) >= 144 THEN ' We are in a VBLANK period!
      REM Here I trigger the vblank interrupt (if enabled via bit 4 of STAT register), and set mode 1
      IF (IoRegisters(&hFF41) AND &b11) <> 1 THEN   ' Check if not already done
        IoRegisters(&hFF41) = IoRegisters(&hFF41) AND &b11111100  ' Here I set
        IoRegisters(&hFF41) = IoRegisters(&hFF41) OR &b00000001   ' mode 1
        IF IoRegisters(&hFF41) AND &b00010000 > 0 THEN   ' If vblank int is enabled, request it
          IoRegisters(&hFF0F) = IoRegisters(&hFF0F) OR &b00000001
        END IF
        DrawBackground()
        DrawSprites()
        RefreshScreen()
      END IF
    ELSE ' Outside the VBlank, perform the MODE 0-2-3 loop
      VideoClkCounterMode += cycles
      IF VideoClkCounterMode <= 201 THEN    ' Just mode 0
        REM Mode 0, do nothing
      ELSEIF VideoClkCounterMode <= 278 THEN                          ' MODE 2 (77..83 clks / 19us) generate OAM interrupt if enabled via bit 5 of STAT (via LCDC int?)
        IF (IoRegisters(&hFF41) AND &b00000011) <> &b10 THEN   '  Check if not already done
          IoRegisters(&hFF41) = IoRegisters(&hFF41) AND &b11111100   ' Here I set
          IoRegisters(&hFF41) = IoRegisters(&hFF41) OR &b00000010    ' mode 2
          IF IoRegisters(&hFF41) AND &b00100000 > 0 THEN   ' If OAM int is enabled, request LCDC int
            IoRegisters(&hFF0F) = IoRegisters(&hFF0F) OR &b00000010
          END IF
        END IF
      ELSEIF VideoClkCounterMode <= 447 THEN                          ' MODE 3 (169..175 clks / 41us)
        'IoRegisters(&hFF41) = IoRegisters(&hFF41) AND &b11111100  ' Here I set
        IoRegisters(&hFF41) = IoRegisters(&hFF41) OR &b00000011    ' mode 3
      ELSE  ' Mode 0 (201..207 clks / 48.6us) and wrap VideoClkCounterMode
        ' Trigger the hblank interrupt if enabled via bit 3 of the stat register (FF41) (via LCDC int?)
        IoRegisters(&hFF41) = IoRegisters(&hFF41) AND &b11111100  ' Here I set mode 0
        VideoClkCounterMode -= 448
        IF IoRegisters(&hFF41) AND &b00001000 > 0 THEN   ' If hblank int is enabled, request LCDC int
          IoRegisters(&hFF0F) = IoRegisters(&hFF0F) OR &b00000010
        END IF
      END IF
  END IF
END SUB
